from discord.components import ComponentType
from discord.ui import Select
from discord import MISSING, Emoji, PartialEmoji
from typing import Callable, Union

from .errors import NotCoroutineError, ComponentTypeError


class Menu:

    def __init__(self, menu_type: ComponentType, selectmenu: "SelectMenu", **kwargs):
        """
        A basic menu from selectMenu class
        """
        self.__menu: Select = Select(select_type=menu_type, **kwargs)
        self.__selectMenu = selectmenu

    def set_callable(self, _callable: Callable) -> "Menu":
        """
        Add a coroutine to the menu
        """
        if not isinstance(_callable, Callable):
            raise NotCoroutineError(_callable)

        self.__selectMenu.set_callable(self.__menu.custom_id, _callable=_callable)
        return self

    def add_option(self, label: str, value: str = MISSING, description: Union[str, None] = None, emoji: Union[str, Emoji, PartialEmoji, None] = None, default: bool = False) -> "Menu":
        """
        Add an option to choice.
        Only from string_select type !
        """
        if self.__menu.type != ComponentType.string_select:
            raise ComponentTypeError()

        self.__menu.add_option(label=label, value=value, description=description, emoji=emoji, default=default)
        return self

    @property
    def component(self) -> Select:
        """
        Get the component
        """
        return self.__menu

    @property
    def callable(self) -> Callable:
        """
        Get the callable link to the menu
        """
        return self.__callable

    @callable.setter
    def callable(self, _callable: Callable):
        """
        Set the callable link to menu
        """


from .typeViews import *
from .views.easy_modified_view import EasyModifiedViews
from .pagination.pagination_view import Pagination
from .multibot.start_multibot import Multibot
from .menu.selectMenu import SelectMenu



from ..views.easy_modified_view import EasyModifiedViews
from .menu import Menu

from typing import Union, Callable
from discord.components import ComponentType
from discord import ChannelType, Member, TextChannel, ApplicationContext

class SelectMenu:
    """
    Create a simply select menu
    """

    def __init__(self, timeout: Union[float, None] = None, disabled_on_timeout: bool = False):
        """
        Init the select menu
        """
        self.__select_menu: EasyModifiedViews = EasyModifiedViews(timeout=timeout, disabled_on_timeout=disabled_on_timeout)

    def add_string_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add a string select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.string_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_user_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add an user select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.user_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_role_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add a role select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.role_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_mentionnable_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add a role select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.mentionable_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_channel_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None, channel_types: list[ChannelType] = None):
        """
        Add a role select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        :param channel_types: A list of channel types that can be selected in this menu.
        """
        return self.__global_add_component(ComponentType.channel_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row, channel_types=channel_types)

    def __global_add_component(self, component_type: ComponentType,
                               custom_id: Union[str, None] = None,
                               placeholder: Union[str, None] = None,
                               min_values: int = 1,
                               max_values: int = 1,
                               disabled: bool = False,
                               row: Union[int, None] = None,
                               channel_types: Union[ChannelType, None] = None) -> Menu:
        """
        global function to add a Select component
        """
        menu = Menu(component_type,
                    self.__select_menu,
                    **{'custom_id': custom_id, 'placeholder': placeholder, 'min_values': min_values,
                       'max_values': max_values, 'disabled': disabled, 'row': row, 'channel_types': channel_types})

        self.__select_menu.add_items(menu.component)

        return menu

    def set_callable(self, *custom_ids: str, _callable : Callable):
        """
        Set a callable for the menu associated with the custom_id
        :param custom_ids: IDs to menus
        """
        self.__select_menu.set_callable(*custom_ids, _callable=_callable)

    async def respond(self, ctx: ApplicationContext, *args, **kwargs) -> None:
        """
        Respond at the ApplicationContext
        """
        await self.__select_menu.respond(ctx=ctx, *args, view=self.__select_menu, **kwargs)

    async def send(self, target: Union[Member, TextChannel], *args, **kwargs) -> None:
        """
        Send at the target
        """
        await self.__select_menu.send(target=target, *args, view=self.__select_menu, **kwargs)

    @property
    def get_view(self) -> EasyModifiedViews:
        """
        Get the current view
        """
        return self.__select_menu



from ..views.easy_modified_view import EasyModifiedViews
from discord import ButtonStyle, Interaction, TextChannel, Member, ApplicationContext
from discord.ui import Button
from .errors import PageNumberNotFound

from typing import Union, Any


class Pagination:
    """
    Class
    -------------
    Allows you to easily setup a view pagination
    """

    def __init__(self, timeout: Union[float, None] = None, disabled_on_timeout: bool = False):
        """
        Initialisation for pagination
        :param timeout: The time before disable items on the view
        :param disabled_on_timeout: If timeout is done, disable all items
        """
        self.__view = EasyModifiedViews(timeout, disabled_on_timeout=disabled_on_timeout)

        self.__view.add_items(Button(label='⏮', row=0, custom_id='back+', style=ButtonStyle.blurple))
        self.__view.add_items(Button(label='◀', row=0, custom_id='back', style=ButtonStyle.blurple))
        self.__view.add_items(Button(label='▶', row=0, custom_id='forward', style=ButtonStyle.blurple))
        self.__view.add_items(Button(label='⏭', row=0, custom_id='forward+', style=ButtonStyle.blurple))
        self.__view.set_callable('back+', 'back', 'forward', 'forward+', _callable=self.__turn_page)

        self.__pages: list[tuple[tuple, dict]] = []
        self.__current_page: int = 0

    def add_page(self, *args, **kwargs) -> None:
        """
        Adds a page (in a list) as if this function directly sent the message
        Pages are just modified and not reset ! Don't forget to disable embeds or content if the page don't need this.

        add_page(content="my message", embeds=[embed1, embed2], ...)
        """
        self.__pages.append((args, kwargs))

    def delete_pages(self, *page_numbers: Union[str, int]):
        """
        Deletes pages in the order in which they were added
        **Start to 0 !**

        delete_pages(0,1,2,3,...)
        """
        nbr_pages = len(self.__pages)-1
        for page_number in page_numbers:

            if page_number < 0 or page_number > nbr_pages:
                raise PageNumberNotFound(page_number)

            del self.__pages[page_number]

    async def __turn_page(self, button, interaction: Interaction):
        """
        Turn the page when button is pressed
        """
        page_count = len(self.__pages)

        if page_count <= 1:
            await self.__view.shutdown()
            await interaction.response.defer(invisible=True)
            return

            # Update the current page based on the button pressed

        if interaction.custom_id == 'back+':  # Go to the first page
            self.__current_page = 0

        elif interaction.custom_id == 'back':  # Go to the previous page
            self.__current_page = max(0, self.__current_page - 1)

        elif interaction.custom_id == 'forward':  # Go to the next page
            self.__current_page = min(page_count-1, self.__current_page + 1)

        elif interaction.custom_id == 'forward+':  # Go to the last page
            self.__current_page = page_count-1


        await interaction.message.edit(

            *self.__pages[self.__current_page][0],

            **self.__pages[self.__current_page][1]

        )

        # Acknowledge the interaction
        await interaction.response.defer(invisible=True)

    async def send(self, target: Union[Member, TextChannel]) -> None:
        """
        Send pagination without introduction message.
        :param target: The member or channel to send the pagination
        """
        await self.__view.send(ctx=target, *self.__pages[0][0], **self.__pages[0][1], view=self.get_view)

    async def respond(self, ctx: ApplicationContext) -> None:
        """
        Respond to the command call
        """
        await self.__view.respond(ctx=ctx, *self.__pages[0][0], **self.__pages[0][1], view=self.get_view)

    @property
    def get_view(self) -> EasyModifiedViews:
        """
        Get the view of pagination
        """
        return self.__view

    @property
    def get_page(self) -> int:
        """
        Get the number of showed page
        """
        return self.__current_page



from discord import Interaction, ApplicationContext, Message, Member, TextChannel
from discord.ui import View, Item
from typing import Union, Callable, Iterable

from ..typeViews import T_views
from .errors import CustomIDNotFound


class EasyModifiedViews(View):
    """
    Class
    -------------
    Allows you to easily modify and replace an ui.
    """

    def __init__(self, timeout: Union[float, None] = None, disabled_on_timeout: bool = False, *items: Item):
        """
        Init a Class view for Discord UI
        :param timeout: The time before ui disable
        """
        super().__init__(*items, timeout=timeout)
        self.__timeout: Union[float, None] = timeout
        self.__disabled_on_timeout: bool = disabled_on_timeout
        self.__callback: dict[str, dict[str, Union[Callable[[Interaction], None], T_views]]] = {}
        self.__ctx: Union[Message, Interaction] = None

    def __check_custom_id(self, custom_id: str) -> None:
        """
        Check if the custom_id is alive
        :param custom_id: ID to find
        :raise: CustomIDNotFound
        """
        if custom_id not in self.__callback.keys():
            raise CustomIDNotFound()

    async def respond(self, ctx: ApplicationContext, *args, **kwargs) -> None:
        """
        Respond at the ApplicationContext
        """
        self.__ctx = await ctx.respond(*args, **kwargs)

    async def send(self, target: Union[Member, TextChannel], *args, **kwargs) -> None:
        """
        Send at the target
        """
        self.__ctx = await target.send(*args, **kwargs)

    def add_items(self,
                   *items: T_views) -> "EasyModifiedViews":
        """
        Add all items in the View.
        """

        for ui in items:

            self.__callback[ui.custom_id] = {'ui': ui, 'func': None}
            self.add_item(ui)

        return self

    async def update_items(self, *items: T_views) -> "EasyModifiedViews":
        """
        Update all views.
        Append items if custom_ids not in the view
        Update items if custom_ids in the view
        :param items: items to update
        """

        for item in items:

            try:
                self.__check_custom_id(item.custom_id)
                self.__callback[item.custom_id]['ui'] = item

            except CustomIDNotFound:
                self.add_items(item)

        await self._update()
        return self

    def set_callable_decorator(self, custom_id: str):
        """
        Decorator to set up a callable for the item

        **Interaction parameter is required in coroutine function !**

        view = EasyModifiedViews(None)
        view.add_view(discord.ui.Button(label='coucou', custom_id='test_ID'))

        @view.set_callable_decorator(custom_id='test_ID')

        async def rep(**UI**, **interaction**):
            await interaction.response.send_message('coucou !!!')

        await ctx.respond('coucou', view=view)

        :param custom_id: item ID of the view
        """

        def decorator(_callable: Callable):
            self.__check_custom_id(custom_id)

            self.__callback[custom_id]['func'] = _callable
            return _callable

        return decorator

    def set_callable(self, *custom_ids: str, _callable: Callable):
        """
        set up a callable for items
        :param custom_ids: items IDs of the view
        :param _callable: The callable linked

        **UI and Interaction parameter is required in callable function !**

        view = EasyModifiedViews(None)

        view.add_view(discord.ui.Button(label='coucou', custom_id='test_ID'))

        async def rep(**UI**, **interaction**):
            await interaction.response.send_message('coucou !!!')

        view.set_callable(custom_id='test_ID', callable=rep)
        await ctx.respond('coucou', view=view)
        """
        for custom_id in custom_ids:
            self.__check_custom_id(custom_id)

            self.__callback[custom_id]['func'] = _callable

    async def interaction_check(self, interaction: Interaction) -> bool:
        """
        Func to apply items
        """
        func = self.__callback[interaction.custom_id]['func']

        if func is not None:
            await func(self.__callback[interaction.custom_id]['ui'], interaction)

        else:
            await interaction.response.defer(invisible=True)

        return True

    async def shutdown(self) -> None:
        """
        Disable all items (ui) in the view
        """
        self.disable_all_items()
        await self._update()

    async def disable_items(self, *custom_ids: str) -> None:
        """
        Disable partial items in the view
        :param custom_ids: custom ids of all items to deactivate
        """

        self.disable_all_items(exclusions=[self.get_item(id_) for id_ in self.__callback.keys() if id_ not in custom_ids])
        await self._update()

    async def enable_items(self, *custom_ids: str) -> None:
        """
        Enabl partial items in the view
        :param custom_ids: custom ids of all items to activate
        """
        self.enable_all_items(
            exclusions=[self.get_item(id_) for id_ in self.__callback.keys() if id_ not in custom_ids])
        await self._update()

    async def full_enable_items(self) -> None:
        """
        Enable all items in the view
        """
        self.enable_all_items()
        await self._update()


    async def switch_status_items(self):
        """
        Switch status for all items
        Enable -> Disable
        Disable -> Enable
        """

        for key, in self.__callback.keys():
            self.__callback[key]['ui'].disabled = not self.__callback[key]['ui'].disabled

        await self._update()


    def is_items_disabled(self, *custom_ids: str) -> bool:
        """
        Return True if all items are disabled
        """

        for custom_id in custom_ids:
            self.__check_custom_id(custom_id)
            if not self.__callback[custom_id]['ui'].disabled:
                return False

        return True


    def is_items_enabled(self, *custom_ids: str) -> bool:
        """
        Return True il aff items are enabled
        """

        for custom_id in custom_ids:
            self.__check_custom_id(custom_id)
            if self.__callback[custom_id]['ui'].disabled:
                return False

        return True


    async def delete_items(self, clear_all: bool = False, *custom_ids: str) -> None:
        """
        Delete an item on the view
        :param custom_ids: IDs of items to delete
        :param clear_all: Clear all items in the view
        :raise: CustomIDNotFound
        """
        if clear_all:
            self.clear_items()

        else:

            for custom_id in custom_ids:
                self.__check_custom_id(custom_id)
                self.remove_item(self.get_item(custom_id))

        await self._update()

    def on_timeout(self) -> None:
        """
        Called if timeout view is finished
        """
        if self.__disabled_on_timeout:
            self.shutdown()

    async def _update(self) -> None:
        """
        Update the View on the attached message.
        """
        try:

            if self.message:
                await self.message.edit(view=self)

            else:
                await self.__ctx.edit(view=self)

        except:
            pass

    @property
    def get_uis(self) -> list[T_views]:
        """
        Get all uis in the view
        """
        return [i['ui'] for i in self.__callback.values()]

    def get_ui(self, custom_id: str) -> T_views:
        """
        Get an ui in the view
        :raise: CustomIDNotFound
        """
        self.__check_custom_id(custom_id)
        return self.__callback[custom_id]['ui']

    def __str__(self):
        return str(self.__callback)